"""uti_plot backend for matplotlib

.. moduleauthor:: N. Canestrari, O. Chubar, R. Nagler, M. Rakitin
"""
import os
import sys
import subprocess
import platform
from array import *
from math import *
import numpy as np
import oasys_srw.uti_plot_com as uti_plot_com
import oasys_srw.uti_math as uti_math
import oasys_srw.uti_plot as uti_plot

class Backend(object):
    def __init__(self, backend, fname_format):
        """Set matplotlib backend depending on IPython state, mpld3 availability and fname_format
        Tries `backend` based on what the execution context is (IPython, X11, etc.).
        Will throw exception if matplotlib or pyplot cannot be imported.

        :param str backend: attempted backend
        :param str fname_format: where to save figs
        """
        import matplotlib
        if self._running_in_ipython():
            backend = self._init_ipython(backend)
            import matplotlib.pyplot
            self._pl = matplotlib.pyplot
        else:
            if backend == uti_plot.DEFAULT_BACKEND:
                (backend, fname_format) = self._default_backend(fname_format)
            matplotlib.use(backend)
            import matplotlib.pyplot
            self._pl = matplotlib.pyplot
            matplotlib.pyplot.ioff()
            (backend, fname_format) = self._verify_pyplot(backend, fname_format)
        if fname_format is not None:
            print('Saving figures to ' + fname_format)
        self._backend = backend
        self._fname_format = fname_format
        self._figure_num = 0
  
    def uti_plot_show(self):
        'Render the plots generated thus far; Prompts user to do the generation'
        self._pyplot_show()

    #def plot1D(ar1d,x_range,labels=('energy, KeV','Ph/s/0.1%BW')):
    def uti_plot1d(self, ar1d, x_range, labels=('Photon Energy [eV]', 'ph/s/0.1%bw')):
        #fig = _pl.figure(figsize=(12,8))
        fig = self._pl.figure()
        self._plot_1D(ar1d,x_range,labels,fig)
        return self._maybe_savefig(fig)

    def uti_plot1d_ir(self, arY, arX, labels=('Longitudinal Position [m]', 'Horizontal Position [m]')): #1D plot on irregular mesh
        fig = self._pl.figure()
        self._plot_1D_ir(arY, arX, labels, fig)
        return self._maybe_savefig(fig)

    def uti_plot1d_m(self, ars, labels=('X', 'Y'), styles=None, legend=None): #OC25102019
        fig = self._pl.figure()
        self._plot_1D_m(ars, labels, fig, styles=styles, legend=legend)
        return self._maybe_savefig(fig)

    #def plot2D(ar2d,x_range,y_range,labels=('Horizontal position, mm','Vertical position, mm')):
    def uti_plot2d(self, ar2d, x_range, y_range, labels=('Horizontal Position [m]','Vertical Position [m]')):
        #fig = _pl.figure(figsize=(12,8))
        fig = self._pl.figure()
        self._plot_2D(ar2d,x_range,y_range,labels,fig)
        return self._maybe_savefig(fig)

    def uti_plot2d1d(self, data, x_range, y_range, xc, yc, labels, _graphs_joined=True):
        x0 = x_range[0]; x1 = x_range[1]; nx = x_range[2]
        #y0 = x_range[0]; y1 = y_range[1]; ny = y_range[2]
        y0 = y_range[0]; y1 = y_range[1]; ny = y_range[2] #OC090714

        label2D = labels[0]; label1H = labels[1]; label1V = labels[2]

        fig = None
        if _graphs_joined:
            #fig = _pl.figure(figsize=(12,5))
            fig = self._pl.figure(figsize=(15,5.3))
           
            self._plot_2D(data, x_range, y_range, label2D, fig, 131) #showing graphs in one panel
        else: self.uti_plot2d(data, x_range, y_range, label2D)

        xStep = 0
        if nx > 1: xStep = (x1 - x0)/(nx - 1)
        yStep = 0
        if ny > 1: yStep = (y1 - y0)/(ny - 1)
        inperpOrd = 1 #interpolation order to use (1 to 3)

        #_plot_1D(data[iy,:],range_x,label1H,fig,132)
        arCutX = array('d', [0]*nx)
        xx = x0
        for ix in range(nx):
            arCutX[ix] = uti_math.interp_2d(xx, yc, x0, xStep, nx, y0, yStep, ny, data, inperpOrd, 1, 0)
            xx += xStep
        if _graphs_joined: self._plot_1D(arCutX, x_range, label1H, fig, 132) #OC150814
        else: self.uti_plot1d(arCutX, x_range, label1H)

        #_plot_1D(data[:,ix],range_y,label1V,fig,133)
        arCutY = array('d', [0]*ny)
        yy = y0
        for iy in range(ny):
            #arCutY[iy] = _interp_2d(xc, yy, x0, xStep, nx, y0, yStep, ny, data, inperpOrd, 1, 0)
            arCutY[iy] = uti_math.interp_2d(xc, yy, x0, xStep, nx, y0, yStep, ny, data, inperpOrd, 1, 0)
            yy += yStep
        if _graphs_joined: self._plot_1D(arCutY, y_range, label1V, fig, 133)
        else: self.uti_plot1d(arCutY, y_range, label1V)

        if _graphs_joined: self._pl.tight_layout() #OC081115

        return self._maybe_savefig(fig)

    #def srw_ascii_plot(fname):
    #def uti_data_file_plot(self, _fname, _read_labels=1, _e=0, _x=0, _y=0, _graphs_joined=1):
    #def uti_data_file_plot(self, _fname, _read_labels=1, _e=0, _x=0, _y=0, _graphs_joined=1, _traj_report=False, _traj_axis='x'): #MR20160729
    #def uti_data_file_plot(self, _fname, _read_labels=1, _e=0, _x=0, _y=0, _graphs_joined=1, _traj_report=False, _traj_axis='x', _scale='linear', _width_pixels=None): #MR27012017
    def uti_plot_data_file(self, _fname, _read_labels=1, _e=0, _x=0, _y=0, _graphs_joined=1, #OC16112017 (renamed: uti_plot_data_file -> uti_plot_data_file)
                           _multicolumn_data=False, _column_x=None, _column_y=None, #MR31102017
                           _scale='linear', _width_pixels=None):
        #data, mode, allrange = srw_ascii_load(fname)
        #data, mode, allrange, arLabels, arUnits = _file_load(_fname, _read_labels)
        #data, mode, allrange, arLabels, arUnits = uti_plot_com.file_load(_fname, _read_labels)
        #data, mode, allrange, arLabels, arUnits = uti_plot_com.file_load(_fname, _read_labels, _traj_report, _traj_axis) #MR20160729

        #data = np.array(data) #MR27012017 (lines added until "pass")
        #if _scale != 'linear':
        #    available_scales = ['log', 'log2', 'log10']
        #    if _scale not in available_scales:
        #        raise ValueError('Scale "{}" is not supported. Available scales: {}.'.format(_scale, ', '.join(available_scales)))
        #    data[np.where(data <= 0.)] = 1.e-23
        #    data = getattr(np, _scale)(data)
        #if _width_pixels:
        #    try:
        #        from scipy.ndimage import zoom
        #        data = np.reshape(data, (allrange[5], allrange[8]), order='f')
        #        resize_factor = float(_width_pixels) / float(allrange[5])
        #        print('Size before: {}  Dimensions: {}'.format(data.size, data.shape))
        #        data = zoom(data, resize_factor)
        #        if _scale == 'linear':
        #            data[np.where(data < 0.)] = 0.0
        #        print('Size after : {}  Dimensions: {}'.format(data.size, data.shape))
        #        allrange = list(allrange)
        #        allrange[5] = data.shape[0]
        #        allrange[8] = data.shape[1]
        #        allrange = tuple(allrange)
        #        data = np.reshape(data, (data.shape[0] * data.shape[1]), order='f')
        #    except:
        #        print('Cannot resize the image - scipy.ndimage.zoom() cannot be imported.')
        #        pass

        data, mode, allrange, arLabels, arUnits = uti_plot_com.file_load(_fname, _read_labels, _multicolumn_data) #MR31102017
        if not _multicolumn_data:
            data = np.array(data)
        #if mode == 3: #OC17112017 (commented-out)
        #    try:
        #        fwhm_dict = uti_math.fwhm(np.linspace(allrange[0], allrange[1], allrange[2]), data, return_as_dict=True)
        #    except:
        #        fwhm_dict = {'fwhm': -1}
        #    print('FWHM: {:.5f} [{}]'.format(fwhm_dict['fwhm'], arUnits[0]))
        if _scale != 'linear':
            available_scales = ['log', 'log2', 'log10']
            if _scale not in available_scales:
                raise ValueError('Scale "{}" is not supported. Available scales: {}.'.format(_scale, ', '.join(available_scales)))
            data[np.where(data <= 0.)] = 1.e-23
            data = getattr(np, _scale)(data)
        if _width_pixels:
            try:
                from scipy.ndimage import zoom
                data = np.reshape(data, (allrange[5], allrange[8]), order='C')
                resize_factor = float(_width_pixels) / float(allrange[5])
                print('Size before: {}  Dimensions: {}'.format(data.size, data.shape))
                data = zoom(data, resize_factor)
                if _scale == 'linear':
                    data[np.where(data < 0.)] = 0.0
                print('Size after : {}  Dimensions: {}'.format(data.size, data.shape))
                allrange = list(allrange)
                allrange[5] = data.shape[0]
                allrange[8] = data.shape[1]
                allrange = tuple(allrange)
                data = np.reshape(data, (data.shape[0] * data.shape[1]), order='C')
            except:
                print('Cannot resize the image - scipy.ndimage.zoom() cannot be imported.')
                pass

        #print(allrange)
        m = self._enum('T','V','H','E','HV','EV','EH','EHV')
        if mode==m.T:
            fig = self.__mode_T(data, allrange, arLabels, arUnits, _e, _x, _y)
        elif mode==m.V:
            fig = self.__mode_V(data, allrange)
        elif mode==m.H:
            fig = self.__mode_H(data, allrange)
        elif mode==m.E:
            fig = self.__mode_E(data, allrange, arLabels, arUnits)
        elif mode==m.HV:
            fig = self.__mode_HV(data, allrange, arLabels, arUnits, _x, _y, _graphs_joined)
        #elif mode==m.EV:
        #  fig = __mode_EV(data,allrange)
        #elif mode==m.EH:
        #  fig = __mode_EH(data,allrange)
        elif mode==m.EHV:
            fig = self.__mode_EHV(data, allrange, arLabels, arUnits, _e, _x, _y, _graphs_joined)

        if _multicolumn_data: #MR31102017
            available_cols = list(data.keys())
            for c in [_column_x, _column_y]:
                if c not in available_cols:
                    raise ValueError('Incorrect column specified: {}.\nAvailable columns: {}'.format(c, available_cols))
            fig = self._plot_1D_XvsY(data, _column_x, _column_y)

        return self._maybe_savefig(fig)

    def uti_data_file_plot(self, _fname, _read_labels=1, _e=0, _x=0, _y=0, _graphs_joined=1,
                           _multicolumn_data=False, _column_x=None, _column_y=None, #MR31102017
                           _scale='linear', _width_pixels=None):
        return self.uti_plot_data_file(_fname, _read_labels, _e, _x, _y, _graphs_joined, _multicolumn_data, _column_x, _column_y, _scale, _width_pixels) #OC16112017

    def _enum(self, *sequential, **named): #Had to copy this in uti_plot_com
        enums = dict(zip(sequential, range(len(sequential))), **named)
        return type('Enum', (), enums)

    def _plot_1D_XvsY(self, data, column_x, column_y, figsize=(8, 5), typ=111): #MR31102017
        fig = self._pl.figure(figsize=figsize)
        ax = fig.add_subplot(typ)
        ax.plot(data[column_x]['data'], data[column_y]['data'])
        ax.grid()
        ax.set_xlabel(data[column_x]['label'])
        ax.set_ylabel(data[column_y]['label'])
        return fig

    def _plot_1D_ir(self, ary, arx, labels, fig, typ=111): #OC15112017
        if isinstance(arx, (list, array)): arx = np.array(arx)
        if isinstance(ary, (list, array)): ary = np.array(ary)

        ax = fig.add_subplot(typ)
        ax.plot(arx, ary)
        ax.grid()
        #ax.set_xlim(arx[0], arx[-1])
        #ax.set_xlim(min(arx), max(arx))

        ax.set_xlabel(labels[0])
        ax.set_ylabel(labels[1])
        if(len(labels) > 2):
            ax.set_title(labels[2])

    def _plot_1D(self, ar1d, x_range, labels, fig, typ=111):
        lenAr1d = len(ar1d)
        if lenAr1d > x_range[2]: ar1d = np.array(ar1d[0:x_range[2]])
        elif lenAr1d < x_range[2]:
            auxAr = array('d', [0]*x_range[2])
            for i in range(lenAr1d): auxAr[i] = ar1d[i]
            ar1d = np.array(auxAr)

        if isinstance(ar1d,(list,array)): ar1d = np.array(ar1d)

        x = np.linspace(x_range[0],x_range[1],x_range[2])
        #print('_plot_1D:', x_range)
        
        ax = fig.add_subplot(typ)
        ax.plot(x,ar1d)
        ax.grid()
        ax.set_xlim(x[0],x[-1])
        ax.set_xlabel(labels[0])
        ax.set_ylabel(labels[1])
        if(len(labels) > 2):
            ax.set_title(labels[2])

    def _plot_1D_m(self, ars, labels, fig, typ=111, styles=None, legend=None): #OC25102019
        #Assumed: ars = [curve_1,curve_2,...], where curve_i can be:
        # curve_i = [[y1_i,y2_i,...],[x_min_i,x_max_i,nx_i]] or
        # curve_i = [[y1_i,y2_i,...],[x1_i,x2_i,...]] or
        # curve_i = [[x1_i,y1_i],[x2_i,y2_i],...] 

        nCurves = len(ars)
        ax = fig.add_subplot(typ)

        legendWasUsed = False
        for i in range(nCurves):
            ar_i = ars[i] #individual curve
            len_ar_i = len(ar_i)

            curStyle = None
            if(styles is not None):
                if(isinstance(styles, (list, array))):
                    if(i < len(styles)): curStyle = styles[i]
            curLegend = None
            if(legend is not None):
                if(isinstance(legend, (list, array))):
                    if(i < len(legend)): curLegend = legend[i]
                    
            arx = None
            ary = None

            if(len_ar_i == 2):
                ar_i_0 = ar_i[0]
                ar_i_1 = ar_i[1]
                len_ar_i_0 = len(ar_i_0)
                len_ar_i_1 = len(ar_i_1)
                #if((len_ar_i_1 == 3) and (len_ar_i_0 == 3)): 
                #    if(ar_i_1[2] == 3): #assuming [[y1,y2,y3],[x_min_i,x_max_i,nx_i]]
                #        ary = ar_i_0
                #        arx = np.linspace(ar_i_1[0], ar_i_1[1], ar_i_1[2])
                #    else: #assuming [[y1,y2,y3],[x1,x2,x3]]
                #        ary = ar_i_0
                #        arx = ar_i_1
                if(len_ar_i_1 == 3): #OC21022019
                    if(len_ar_i_0 == 3): 
                        if(ar_i_1[2] == 3): #assuming [[y1,y2,y3],[x_min_i,x_max_i,nx_i]]
                            ary = ar_i_0
                            arx = np.linspace(ar_i_1[0], ar_i_1[1], ar_i_1[2])
                        else: #assuming [[y1,y2,y3],[x1,x2,x3]]
                            ary = ar_i_0
                            arx = ar_i_1
                    elif(len_ar_i_0 == ar_i_1[2]): #assuming [[y1,y2,y3,y4,...],[x_min_i,x_max_i,nx_i]]
                        ary = ar_i_0
                        arx = np.linspace(ar_i_1[0], ar_i_1[1], ar_i_1[2])
                    
                elif((len_ar_i_1 == 2) and (len_ar_i_0 == 2)): #assuming [[x1,y1],[x2,y2]]
                    arx = [ar_i_0[0],ar_i_1[0]]
                    ary = [ar_i_0[1],ar_i_1[1]]

                elif((len_ar_i_1 != 1) and (len_ar_i_1 == len_ar_i_1)): #assuming [[y1,y2,...],[x1,x2,...]]
                    ary = ar_i_0
                    arx = ar_i_1

            elif(len_ar_i > 2): #assuming [[x1_i,y1_i],[x2_i,y2_i],...]
                arx = [0]*len_ar_i
                ary = [0]*len_ar_i
                for j in range(len_ar_i):
                    ar_ij = ar_i[j]
                    arx[j] = ar_ij[0]
                    ary[j] = ar_ij[1]
                    
            if((curStyle is None) and (curLegend is None)): ax.plot(arx, ary)
            elif((curStyle is not None) and (curLegend is None)): ax.plot(arx, ary, curStyle)
            elif((curStyle is None) and (curLegend is not None)): ax.plot(arx, ary, label=curLegend)
            elif((curStyle is not None) and (curLegend is not None)): ax.plot(arx, ary, curStyle, label=curLegend)

            if(curLegend is not None): legendWasUsed = True

        if(legendWasUsed): ax.legend()
        ax.grid()
        ax.set_xlabel(labels[0])
        ax.set_ylabel(labels[1])
        if(len(labels) > 2): ax.set_title(labels[2])

    def _plot_2D(self, ar2d, x_range, y_range, labels, fig, typ=111):

        #totLen = int(x_range[2]*y_range[2])
        #lenAr2d = len(ar2d)
        #if lenAr2d > totLen: ar2d = np.array(ar2d[0:totLen])
        #elif lenAr2d < totLen:
        #    auxAr = array('d', [0]*lenAr2d)
        #    for i in range(lenAr2d): auxAr[i] = ar2d[i]
        #    ar2d = np.array(auxAr)

        #if isinstance(ar2d,(list,array)): ar2d = np.array(ar2d)
        #ar2d = ar2d.reshape(y_range[2],x_range[2])

        isDataAr = False  #OC30052020
        if isinstance(ar2d, (list, array)):
            isFlat = True
            if isinstance(ar2d[0], (list, array)): isFlat = False

            if((x_range is None) or (y_range is None)):
                if isFlat: raise Exception('Mesh / grid description for 2D plot is not provided')
                if x_range is None:
                    nx = len(ar2d[0])
                    x_range = [0, nx - 1, nx]
                if y_range is None:
                    ny = len(ar2d)
                    y_range = [0, ny - 1, ny]

            if isFlat:
                totLen = int(x_range[2]*y_range[2])
                lenAr2d = len(ar2d)
                if lenAr2d > totLen: ar2d = np.array(ar2d[0:totLen])
                elif lenAr2d < totLen:
                    auxAr = array('d', [0]*lenAr2d)
                    for i in range(lenAr2d): auxAr[i] = ar2d[i]
                    ar2d = np.array(auxAr)
            
            ar2d = np.array(ar2d)
            ar2d = ar2d.reshape(y_range[2], x_range[2])
            isDataAr = True
        else: #Perhaps this is not necessary? Image will be displayed in standard way then (with pixel row number starting from top)
            nx = 0; ny = 0
            if(x_range is None):
                nx, ny = ar2d.size
                x_range = [0, nx - 1, nx]
            if(y_range is None):
                if(ny == 0): nx, ny = ar2d.size
                y_range = [0, ny - 1, ny]

        ax = fig.add_subplot(typ)
           
        #x = np.linspace(x_range[0],x_range[1],x_range[2])
        #y = np.linspace(y_range[0],y_range[1],y_range[2])
        x = np.linspace(x_range[0],x_range[1],x_range[2]) if x_range is not None else None #OC30052020
        y = np.linspace(y_range[0],y_range[1],y_range[2]) if y_range is not None else None 
        
        #ax.pcolormesh(x,y,ar2d,cmap=self._pl.cm.Greys_r) #OC150814
        if isDataAr: ax.pcolormesh(x, y, ar2d, cmap=self._pl.cm.Greys_r) #OC30052020
        else: ax.imshow(ar2d, cmap=self._pl.cm.Greys_r) #OC30052020 (assuming PIL.Image)
        
        #ax.set_xlim(x[0],x[-1])
        #ax.set_ylim(y[0],y[-1])
        if x is not None: ax.set_xlim(x[0],x[-1]) #OC30052020
        if y is not None: ax.set_ylim(y[0],y[-1])
        
        ax.set_xlabel(labels[0])
        ax.set_ylabel(labels[1])
        if(len(labels) > 2): ax.set_title(labels[2])

    def __mode_T(self, data, allrange, _ar_labels, _ar_units, _ec=0, _xc=0, _yc=0):
        #allrange, units = _rescale_range(allrange)
        #allrange, units = _rescale_range(allrange, _ar_units, _ec, _xc, _yc)
        allrange, units = uti_plot_com.rescale_range(allrange, _ar_units, _ec, _xc, _yc)

        #e0, e1, ne, x0, x1, nx, y0, y1, ny = allrange
        e0, e1, ne, x0, x1, nx, y0, y1, ny, ec, xc, yc = allrange
        #toprint = (e0,e1,units[0], x0,x1,units[1], y0,y1,units[2], data[0]) #squeeze have reduced it to an array with one element.
        toprint = (e0,units[0], x0,units[1], y0,units[2], data[0],units[3]) #squeeze have reduced it to an array with one element.
        #sys.stdout.write('Total Flux for \nE: %f -> %f %s\nX: %f -> %f %s\nY: %f -> %f %s\n is %f Ph/s/0.1%BW' % toprint) 
        sys.stdout.write(_ar_labels[3] + ' for \n' + _ar_labels[0] + ': %f %s\n' + _ar_labels[1] + ': %f %s\n' + _ar_labels[2] + ': %f %s\n is %f %s' % toprint) 
        return None

    def __mode_V(self, data, allrange, _ar_labels, _ar_units):
        #allrange, units = _rescale_range(allrange)
        #allrange, units = _rescale_range(allrange, _ar_units)
        allrange, units = uti_plot_com.rescale_range(allrange, _ar_units)

        #e0, e1, ne, x0, x1, nx, y0, y1, ny = allrange
        e0, e1, ne, x0, x1, nx, y0, y1, ny, ec, xc, yc = allrange
        range_y = y0, y1, ny
        #label = ("Vertical Position, ["+units[2]+"]","Ph/s/0.1%BW/mm^2")
        label = (_ar_labels[2] + ' [' + units[2] + ']', _ar_labels[3] + ' [' + _ar_units[3] + ']')
        fig = self._pl.figure(figsize=(4,4))
        self._plot_1D(data, range_y, label, fig)
        return fig

    def __mode_H(self, data, allrange, _ar_labels, _ar_units):
        #allrange, units = _rescale_range(allrange)
        #allrange, units = _rescale_range(allrange, _ar_units)
        allrange, units = uti_plot_com.rescale_range(allrange, _ar_units)

        #e0, e1, ne, x0, x1, nx, y0, y1, ny = allrange
        e0, e1, ne, x0, x1, nx, y0, y1, ny, ec, xc, yc = allrange
        range_x = x0, x1, nx 
        #label = ("Horizontal Position, ["+units[1]+"]","Ph/s/0.1%BW/mm^2")
        label = (_ar_labels[1] + ' [' + units[1] + ']', _ar_labels[3] + ' [' + _ar_units[3] + ']')
        fig = self._pl.figure(figsize=(4,4))
        self._plot_1D(data, range_x, label, fig)
        return fig

    def __mode_E(self, data, allrange, _ar_labels, _ar_units):
        #allrange, units = _rescale_range(allrange)
        #allrange, units = _rescale_range(allrange, _ar_units)
        allrange, units = uti_plot_com.rescale_range(allrange, _ar_units)

        e0, e1, ne, x0, x1, nx, y0, y1, ny, ec, xc, yc = allrange
        range_e = e0, e1, ne
        #label = ("Energy, ["+units[0]+"]","Ph/s/0.1%BW")
        label = (_ar_labels[0] + ' [' + units[0] + ']', _ar_labels[3] + ' [' + _ar_units[3] + ']')
        #fig = self._pl.figure(figsize=(4,4))
        fig = self._pl.figure(figsize=(7,4)) #OC17112017
        self._plot_1D(data,range_e,label,fig)
        return fig

    def __mode_HV(self, data, allrange, _ar_labels, _ar_units, _xc=0, _yc=0, _graphs_joined=True):
        #Could be moved to uti_plot_com.py (since there is not Matplotlib-specific content)
        #allrange, units = _rescale_range(allrange)
        #allrange, units = _rescale_range(allrange, _ar_units, 0, _xc, _yc)
        allrange, units = uti_plot_com.rescale_range(allrange, _ar_units, 0, _xc, _yc)

        e0, e1, ne, x0, x1, nx, y0, y1, ny, ec, xc, yc = allrange
        range_x = x0, x1, nx
        range_y = y0, y1, ny
        #print range_x, range_y
        #x = np.linspace(x0, x1, nx)
        #y = np.linspace(y0, y1, ny)
        #ix = np.where(abs(x)==abs(x).min())[0][0]
        #iy = np.where(abs(y)==abs(y).min())[0][0]
        #label2D = ("Horizontal Position, ["+units[1]+"]", "Vertical Position, ["+units[2]+"]")
        strTitle = _ar_labels[3]
        if (ne == 1) and (e0 > 0): strTitle += ' at ' + str(e0) + ' ' + units[0]

        label2D = (_ar_labels[1] + ' [' + units[1]+ ']', _ar_labels[2] + ' [' + units[2] + ']', strTitle)
        #print(label2D)

        #label1H = ("Horizontal Position, ["+units[1]+"]","Ph/s/0.1%BW/mm^2")
        strTitle = 'At ' + _ar_labels[2] + ': ' + str(yc)
        if yc != 0: strTitle += ' ' + units[2]
        label1H = (_ar_labels[1] + ' [' + units[1] + ']', _ar_labels[3] + ' [' + _ar_units[3] + ']', strTitle)

        #label1V = ("Vertical Position, ["+units[2]+"]","Ph/s/0.1%BW/mm^2")
        strTitle = 'At ' + _ar_labels[1] + ': ' + str(xc)
        if xc != 0: strTitle += ' ' + units[1]
        label1V = (_ar_labels[2] + ' [' + units[2] + ']', _ar_labels[3] + ' [' + _ar_units[3] + ']', strTitle)

        #return plot_2D_1D(data, range_x, range_y, [label2D, label1H, label1V], _graphs_joined)
        return self.uti_plot2d1d(data, range_x, range_y, xc, yc, [label2D, label1H, label1V], _graphs_joined)

    ##    fig = None
    ##    if _graphs_joined:
    ##        fig = _pl.figure(figsize=(12,5))
    ##        _plot_2D(data, range_x, range_y, label2D, fig, 131) #showing graphs in one figure
    ##    else: uti_plot2d(data, range_x, range_y, label2D)
    ##
    ##    xStep = 0
    ##    if nx > 1: xStep = (x1 - x0)/(nx - 1)
    ##    yStep = 0
    ##    if ny > 1: yStep = (y1 - y0)/(ny - 1)
    ##    inperpOrd = 1 #interpolation order (1 to 3)
    ##
    ##    #_plot_1D(data[iy,:],range_x,label1H,fig,132)
    ##    arCutX = array('d', [0]*nx)
    ##    xx = x0
    ##    for ix in range(nx):
    ##        #arCutX[ix] = _interp_2d(xx, yc, x0, xStep, nx, y0, yStep, ny, data, inperpOrd, 1, 0)
    ##        arCutX[ix] = uti_math.interp_2d(xx, yc, x0, xStep, nx, y0, yStep, ny, data, inperpOrd, 1, 0)
    ##        xx += xStep
    ##    if _graphs_joined: _plot_1D(arCutX, range_x, label1H, fig, 132)
    ##    else: uti_plot1d(arCutX, range_x, label1H)
    ##
    ##    #_plot_1D(data[:,ix],range_y,label1V,fig,133)
    ##    arCutY = array('d', [0]*ny)
    ##    yy = y0
    ##    for iy in range(ny):
    ##        #arCutY[iy] = _interp_2d(xc, yy, x0, xStep, nx, y0, yStep, ny, data, inperpOrd, 1, 0)
    ##        arCutY[iy] = uti_math.interp_2d(xc, yy, x0, xStep, nx, y0, yStep, ny, data, inperpOrd, 1, 0)
    ##        yy += yStep
    ##    if _graphs_joined: _plot_1D(arCutY, range_y, label1V, fig, 133)
    ##    else: uti_plot1d(arCutY, range_y, label1V)
    ##
    ##    return fig

    #def __mode_EV(data, allrange):
    #  #to be updated
    #  allrange, units = _rescale_range(allrange)
    #  e0, e1, ne, x0, x1, nx, y0, y1, ny = allrange
    #  range_e = e0, e1, ne  
    #  range_y = y0, y1, ny
    #  e = np.linspace(e0, e1, ne)
    #  y = np.linspace(y0, y1, ny)
    #  ie = np.where(data.sum(axis=1)==data.sum(axis=1).max())[0][0]
    #  iy = np.where(abs(y).min())[0][0]
    #  label1E = ("Energy, ["+units[0]+"]","Ph/s/0.1%BW/mm^2")
    #  label1V = ("Vertical Position, ["+units[2]+"]","Ph/s/0.1%BW/mm^2")
    #  fig = _pl.figure(figsize=(8,4))
    #  _plot_1D(data[iy,:],range_e,label1E,fig,121)
    #  _plot_1D(data[:,ie],range_y,label1V,fig,122)
    #  return fig

    #def __mode_EH(data, allrange):
    #  #to be updated
    #  allrange, units = _rescale_range(allrange)
    #  e0, e1, ne, x0, x1, nx, y0, y1, ny = allrange
    #  range_e = e0, e1, ne  
    #  range_x = x0, x1, nx
    #  e = np.linspace(e0, e1, ne)
    #  x = np.linspace(x0, x1, nx)
    #  ie = np.where(data.sum(axis=1)==data.sum(axis=1).max())[0][0]
    #  ix = np.where(abs(y).min())[0][0]
    #  label1E = ("Energy, ["+units[0]+"]","Ph/s/0.1%BW/mm^2")
    #  label1H = ("Horizontal Position, ["+units[1]+"]","Ph/s/0.1%BW/mm^2")
    #  fig = _pl.figure(figsize=(8,4))
    #  _plot_1D(data[ix,:],range_e,label1E,fig,121)
    #  _plot_1D(data[:,ie],range_x,label1H,fig,122)
    #  return fig

    def __mode_EHV(self, data, allrange, _ar_labels, _ar_units, _ec, _xc, _yc, _graphs_joined=1):
        #allrange, units = _rescale_range(allrange)
        #allrange, units = _rescale_range(allrange, _ar_units, _ec, _xc, _yc)
        allrange, units = uti_plot_com.rescale_range(allrange, _ar_units, _ec, _xc, _yc)

        #e0, e1, ne, x0, x1, nx, y0, y1, ny = allrange
        e0, e1, ne, x0, x1, nx, y0, y1, ny, ec, xc, yc = allrange

        #e = np.linspace(e0, e1, ne)
        #x = np.linspace(x0, x1, nx)
        #y = np.linspace(y0, y1, ny)
        #ie = np.where(data.sum(axis=1)==data.sum(axis=1).max())[0][0]
        #ix = np.where(abs(x)==abs(x).min())[0][0]
        #iy = np.where(abs(y)==abs(y).min())[0][0]

        range_e = e0, e1, ne
        range_x = x0, x1, nx
        range_y = y0, y1, ny

        ie = 0
        if ne > 1:
            if ec > e1: ie = ne - 1
            elif ec > e0:
                eStep = (e1 - e0)/(ne - 1)
                if eStep > 0: ie = int(round((ec - e0)/eStep))
        ix = 0
        if nx > 1:
            if xc > x1: ix = nx - 1
            elif xc > x0:
                xStep = (x1 - x0)/(nx - 1)
                if xStep > 0: ix = int(round((xc - x0)/xStep))
        iy = 0
        if ny > 1:
            if yc > y1: iy = ny - 1
            elif yc > y0:
                yStep = (y1 - y0)/(ny - 1)
                if yStep > 0: iy = int(round((yc - y0)/yStep))

        #label2D = ("Horizontal Position, ["+units[1]+"]", "Vertical Position, ["+units[2]+"]")
        label2D = (_ar_labels[1] + ' [' + units[1] + ']', _ar_labels[2] + ' [' + units[2] + ']')

        #label1E = ("Energy, ["+units[0]+"]","Ph/s/0.1%BW/mm^2")
        label1E = (_ar_labels[0] + ' [' + units[0] + ']', _ar_labels[3] + ' [' + units[3] + ']')

        #label1H = ("Horizontal Position, ["+units[1]+"]","Ph/s/0.1%BW/mm^2")
        label1H = (_ar_labels[1] + ' [' + units[1] + ']', _ar_labels[3] + ' [' + units[3] + ']')

        #label1V = ("Vertical Position, ["+units[2]+"]","Ph/s/0.1%BW/mm^2")
        label1V = (_ar_labels[2] + ' [' + units[2] + ']', _ar_labels[3] + ' [' + units[3] + ']')

        arCutXY = array('d', [0]*nx*ny)
        perY = ne*nx
        i = 0
        for iiy in range(ny):
            perY_iiy = perY*iiy
            for iix in range(nx):
                arCutXY[i] = data[ie + ne*iix + perY_iiy]
                i += 1

        arCutE = array('d', [0]*ne)
        perX_ix = ne*ix
        perY_iy = perY*iy
        for iie in range(ne): arCutE[iie] = data[iie + perX_ix + perY_iy]

        arCutX = array('d', [0]*nx)
        for iix in range(nx): arCutX[iix] = data[ie + ne*iix + perY_iy]

        arCutY = array('d', [0]*ny)
        for iiy in range(ny): arCutY[iiy] = data[ie + perX_ix + perY*iiy]

        #fig = _pl.figure(figsize=(8,8))
        #_plot_2D(data[:,:,ie], range_x, range_y, label2D, fig, 221)
        #_plot_1D(data[ix,iy,:],range_e,label1E,fig,224)
        #_plot_1D(data[ie,:,iy],range_x,label1H,fig,222)
        #_plot_1D(data[:,ix,ie],range_y,label1V,fig,223)

        fig = None
        if _graphs_joined:
            fig = self._pl.figure(figsize=(12,5))
            self._plot_2D(arCutXY, range_x, range_y, label2D, fig, 221) #showing graphs in one figure
            self._plot_1D(arCutE, range_e, label1E, fig, 222)
            self._plot_1D(arCutX, range_x, label1X, fig, 223)
            self._plot_1D(arCutY, range_y, label1Y, fig, 224)
        else:
            self.uti_plot2d(arCutXY, range_x, range_y, label2D)
            self.uti_plot1d(arCutE, range_e, label1E)
            self.uti_plot1d(arCutX, range_x, label1X)
            self.uti_plot1d(arCutY, range_y, label1Y)
        return self._maybe_savefig(fig)

    def _maybe_savefig(self, figure):
        """If there is an ``fname_format``, then save the pyplot `figure`
        (if not None) to the computed file name
        """
        if figure is None:
            return figure
        if self._fname_format is None:
            return figure
        figure.savefig(self._fname_format.format(self._figure_num))
        self._figure_num += 1
        return figure

    def _pyplot_show(self):
        """'Render the plots, if the backend is interactive.
        Plots saved to files (``fname_format``), have already been written.
        """
        with self._HideGUIErrorOutput():
            self._pl.show()

    def _default_fname_format(self):
        """Use main program's name (or uti_plot if none) for default file format name.
        The default file type is png
        """
        try:
            base = os.path.splitext(
                os.path.basename(sys.modules['__main__']['__file__']))
        except:
            base = 'uti_plot'
        fname_format = base + '-{}.png'
        return fname_format

    def _default_backend(self, fname_format):
        """Selects backend based on execution context (X11, macosx, etc.).  If an
        interactive backend is not possible, return _default_file_backend()
        """
        if self._running_in_x11() or self._running_in_windows():
            backend = 'TkAgg'
        elif self._running_in_macosx():
            backend = 'macosx'
        else:
            (backend, fname_format) = self._default_file_backend(fname_format)
        return (backend, fname_format)

    def _default_file_backend(self, fname_format):
        """Returns "agg" for backend with fname_format.  If fname_format is none,
        calls _default_fname_format() to compute fname_format.
        """
        if fname_format is None:
            fname_format = self._default_fname_format()
        return ('agg', fname_format)

    def _init_ipython(self, backend):
        """Tries `backend`.  Returns ``inline`` backend if default requested or
        unable to load requested backend.
        """
        if backend == uti_plot.DEFAULT_BACKEND:
            backend = 'inline'
        is_mpld3 = backend == 'mpld3'
        b = 'inline' if is_mpld3 else backend
        get_ipython().magic('matplotlib ' + b)
        if is_mpld3:
            try:
                import mpld3
                mpld3.enable_notebook()
            except:
                print('mpld3: backend unavailable; plots will be inline')
                backend = 'inline'
        return backend

    def _verify_pyplot(self, backend, fname_format):
        """Create a pyplot figure and close it to be sure backend functions properly.
        If backend fails, configures _default_file_backend().
        """
        import matplotlib.pyplot
        pl = matplotlib.pyplot
        try:
            pl.figure(figsize=(1,1)) #OC08032020 (the line below throws exception with ~recent matplotlib)
            #pl.figure(figsize=(0,0))
            pl.close('all')
        except:
        #except Exception as e:
            #print(str(e))
            old = backend
            (backend, fname_format) = self._default_file_backend(fname_format)
            pl.switch_backend(backend)
            # If this raises an exception, uti_plot will set backend to None 
            pl.figure()
            pl.close('all')
            print(old + ': backend unavailable; plots will be saved to files')
        return (backend, fname_format)

    def _running_in_ipython(self):
        'Is the current interpreter IPython?'
        try:
            __IPYTHON__
            return True
        except NameError:
            return False

    def _running_in_macosx(self):
        'Is Mac OS X running?'
        return platform.system() == 'Darwin'

    def _running_in_windows(self):
        'Is Darwin running?'
        return platform.system() == 'Windows'

    def _running_in_x11(self):
        'Is X11 running?'
        d = os.environ.get('DISPLAY')
        if d is None or d == '':
            return False
        devnull = open(os.devnull, 'wb')
        tries = [
            # xset is not in core install on some OSes
            ['xset', '-q'],
            # This takes a couple of seconds
            ['xterm', '/bin/true']]
        for t in tries:
            try:
                subprocess.call(t, stdout=devnull, stderr=devnull)
                return True
            except:
                pass
        #return false
        return False #MR19042016

    class _HideGUIErrorOutput(object):
        """Redirect low level stderr (fd #2) to os.devnull as context manager

        TkAgg has a bug::

            can't invoke "event" command:  application has been destroyed
                while executing
            "event generate $w <<ThemeChanged>>"
                (procedure "ttk::ThemeChanged" line 6)
                invoked from within
            "ttk::ThemeChanged"

        The output comes from C so it has to be redirected at the file descriptor
        level.  This class does the low level redirection so that message doesn't
        come out.  Example::

            with HideGUIErrorOutput():
                pl.show()
        """
        def __init__(self):
            pass

        def __enter__(self):
            'Redirects stderr to devnull, saving _prev_stderr'
            if sys.stderr is None or sys.__stderr__ is None:
                return
            sys.stderr.flush()
            fno = sys.__stderr__.fileno() #OC150814
            if(fno >= 0):
                null = os.open(os.devnull, os.O_WRONLY)
                #self._prev_stderr = os.dup(sys.__stderr__.fileno())
                self._prev_stderr = os.dup(fno)
                os.dup2(null, sys.__stderr__.fileno())
                os.close(null)

        def __exit__(self, exc_type, exc_value, traceback):
            'Redirects to _prev_stderr'
            if sys.stderr is None or sys.__stderr__ is None:
                return
            sys.__stderr__.flush()
            fno = sys.__stderr__.fileno() #OC150814
            if(fno >= 0):
                #os.dup2(self._prev_stderr, sys.__stderr__.fileno())
                os.dup2(self._prev_stderr, fno)
                os.close(self._prev_stderr)



#############################################################################
#This is a subset of Gianluca's and Ilya's module to read-in Genesis output files
#and to generate SRW wavefront structure
#############################################################################

#from mpi4py import MPI
#comm  = MPI.COMM_WORLD   #Creates one common object "communicator" knwoning about processes
#rank  = comm.Get_rank()  #Process number
#nproc = comm.Get_size()  #Total number of processes

from __future__ import print_function #Python 2.7 compatibility
from oasys_srw.srwlib import *

#from xframework.adaptors.genesis import *
##################from ocelot.adaptors.genesis import *

#from prop_lines import *
import copy
import struct

#from scipy import interpolate

##################################################################################################################
############################################ ROUTINES FROM OCELOT ################################################
####                                                                                                          ####
#### These are to read the Genesis Output file; necessary here in order to make the module 'standalone' ##########
#######                                          i.e. without Ocelot                                    ##########
##################################################################################################################              

#import numpy as np
#from copy import copy, deepcopy

class GenesisOutput:
    
    def __init__(self):
        self.z = []
        self.I = []
        self.n = []
        self.zSlice = []
        self.E = []
        self.aw = []
        self.qfld = []
        
        self.sliceKeys = []
        self.sliceValues = {}
        
        self.parameters = {}
    
    def __call__(self, name):
        '''
        if name not in self.__dict__.keys():
            return 0
        else:
            return self.__dict__[name]
        '''
        if name not in self.parameters.keys():
            return 0.0
        else:
            p, = self.parameters[name]
            return float(p.replace('D','E'))
        

#******************************************************************************
def uti_io_read_Genesis_output(fileName, _short=True):
#def readGenesisOutput(fileName, ):

    out = GenesisOutput()
    
    chunk = 'header'
    
    #nSlice = 0
    
    f=open(fileName,'r')
    f.readline()
    for line in f:

        #print(line)
        
        #tokens = line.replace('=', ' ').split()
        tokens = line.strip().split()

        if len(tokens) < 1:
            #chunk = 'none'
            continue
        
        if tokens == ['z[m]', 'aw', 'qfld']:

            if(_short): return out #OC
            
            chunk = 'optics'
            #print 'reading optics '
            print('reading optics')
            continue
        
        if tokens[0] == 'Input':
            chunk = 'input'
            #print 'reading input parameters'
            #print('reading input parameters')
            continue
        
        #********** output: slice    10
        if tokens[0] == '**********':
            #print 'slice:', tokens[3]
            chunk = 'slices'
            nSlice = int(tokens[3])
         
        if tokens[0] == 'power':
            chunk = 'slice'
            out.sliceKeys = copy.copy(tokens)
            #out.sliceValues[nSlice] = copy.copy(tokens)
            out.sliceValues[nSlice] = {}
            for i in range(0,len(tokens)):
                out.sliceValues[nSlice][out.sliceKeys[i]] = []
            #print 'reading slices'
            #print out.sliceKeys
            #print out.sliceValues[nSlice]
            continue
            
        if chunk == 'optics':
            z,aw,qfld = map(float,tokens)
            out.z.append(z)
            out.aw.append(aw)
            out.qfld.append(qfld)
            
        if chunk == 'input':
            #tokens=line.replace('=','').strip().split()
            tokens=line.replace('=',' ').strip().split() #OC: fix to ensure correct parsing of tokens like 'nslice=10932'
            out.parameters[tokens[0]] = tokens[1:]
            #print('input:', tokens)

        if chunk == 'slice':

            vals = map(float,tokens)
            if sys.hexversion >= 0x030000F0: vals = list(vals) #OC: for compatibility with Py3
            
            #print vals
            for i in range(0,len(vals)):
                out.sliceValues[nSlice][out.sliceKeys[i]].append(vals[i])
            
            #out.zSlice.append(vals[2])
            #out.aw.append(aw)
            #out.qfld.append(qfld)

        if chunk == 'slices':
            if len(tokens) == 2 and tokens[1]=='current':
                #print tokens[1]
                out.I.append(float(tokens[0]))
                out.n.append(nSlice)

    #out.nSlices = int(out('nslice'))
    #out.nSlices = len(out.sliceValues) 
    #out.nZ = len(out.sliceValues[1][out.sliceKeys[0]])

    #print 'nSlice', out.nSlices
    #print 'nZ', out.nZ
    #print('nSlice', out.nSlices)
    #print('nZ', out.nZ)

    return out

#******************************************************************************
#def uti_io_read_Genesis_rad(fileName='simulation.gout.dfl', npoints=51, slice_start=0, slice_end = -1, idx=None):
def uti_io_read_Genesis_rad(fileName='simulation.gout.dfl', _ncar=51, _mult=1., _slice_skip_per=0, _slice_start=0, _slice_end=-1, _align_out='txy'):
    
    #def read_in_chunks(_file, _size=1024):
    #    while True:
    #        data = _file.read(_size)
    #        if not data:
    #            break
    #        yield data

    #def read_in_data_block(_file, _ofst=0, _size=1024):
    #    if(_ofst > 0): _file.seek(_ofst)
    #    return _file.read(_size)

    ncarE2 = _ncar*_ncar
    slice_size = int(ncarE2*16)

    f = open(fileName,'rb')
    
    f.seek(0,2)
    #total_data_len = f.tell() / 8 / 2
    total_data_len = f.tell() / 16 # = _ncar*_ncar*n_slice

    nSliceTot = int(round(total_data_len/ncarE2))
    iSliceEnd = int(_slice_end)
    if(_slice_end < 0): iSliceEnd = nSliceTot - 1

    nSliceToRead = int((iSliceEnd + 1 - _slice_start)/(_slice_skip_per + 1))
    nResVal = nSliceToRead*ncarE2*2
    
    #print('starting array allocation...')
    resFldData = srwl_uti_array_alloc('f', nResVal)
    #print('done')

    #if _slice_start > 0:
    #    f.seek(slice_size*slice_start,0)

    #if slice_end > 0:
    #    data_size = (slice_end - slice_start) * slice_size
    #else:
    #    data_size = total_data_len * 16 - (slice_start) * slice_size

    #print('slice size = ', slice_size)
    #print('data size = ', data_size)
    #print('total_data_len = ', total_data_len)
    
    #ncar = int(np.sqrt((slice_size/16)))
    #print('ncar=', ncar)
    
    #n_slices = total_data_len / (slice_size/16)
    
    #if idx == None:
        #slices = np.zeros([n_slices,ncar,ncar], dtype=complex)
        #slices = array('f', [0]*(n_slices*ncar*ncar*2))
        #slices = srwl_uti_array_alloc('f', (total_data_len*2))
    #else:
        #slices = np.zeros([len(idx),ncar,ncar], dtype=complex)
        #slices = array('f', [0]*(len(idx)*ncar*ncar*2))
        #slices = srwl_uti_array_alloc('f', (len(idx)*ncarE2*2))

    perT = 2
    perX = nSliceToRead*perT
    perY = perX*_ncar

    if((_align_out == 'xyt') or (_align_out == 'xye')):
        perX = 2
        perY = perX*_ncar
        perT = perY*_ncar
    elif((_align_out == 'yxt') or (_align_out == 'yxe')):
        perY = 2
        perX = perY*_ncar
        perT = perX*_ncar
    #else...

    nSliceToRead_mi1 = nSliceToRead - 1

    ofstFile = _slice_start*slice_size
    ofstFileStep = _slice_skip_per*slice_size
    f.seek(ofstFile, 0)

    for iSlice in range(nSliceToRead):
        piece = f.read(slice_size)
        iSlice_perT = iSlice*perT
        i = 0
        for iy in range(_ncar):
            iy_perY = iy*perY
            for ix in range(_ncar):
                ofstRes = int(iSlice_perT + ix*perX + iy_perY)
                i16 = i*16
                i16p8 = i16 + 8
                resFldData[ofstRes] = _mult*(struct.unpack('d', piece[i16 : i16p8])[0])
                resFldData[ofstRes + 1] = _mult*(struct.unpack('d', piece[i16p8 : i16p8 + 8])[0])
                i += 1
        if(_slice_skip_per > 0):
            if(iSlice < nSliceToRead_mi1): f.seek(ofstFileStep, 1) #move to next read position

        #print('slices read:', iSlice)
    return resFldData

#******************************************************************************
def uti_io_set_wfr_from_Genesis_files(_fpGenOut, _fpGenRad, _pol='h', _slice_skip_per=0, _slice_start=0, _slice_end=-1):
    """Sets up SRW wavefront from Genesis output
    :param _fpGenOut: file path to Genesis Output file
    :param _fpGenRad: file path to Genesis Radiation file
    :param _pol: polarization ('h' for horizontal, 'v' for vertical,...)
    :param _slice_skip_per: number of consequtive slice to skip after reading-in one slice
    :param _slice_start: first slice number (in the Genesis file) to read in
    :param _slice_end: last slice number (in the Genesis file) to read in
    """

    outGenesis = uti_io_read_Genesis_output(_fpGenOut)
    
    ncar = int(outGenesis('ncar'))
    zrayl = outGenesis('zrayl')
    rmax0 = outGenesis('rmax0')
    xlamds = outGenesis('xlamds')
    rxbeam = outGenesis('rxbeam')
    rybeam = outGenesis('rybeam')
    zsep = outGenesis('zsep')
    ntail = outGenesis('ntail')
    itdp = int(outGenesis('itdp'))
    xlamd = outGenesis('xlamd')

    pi = 3.14159265358979
    lightSpeed = 2.9979245812E+08 #Speed of Light [m/c]
    light_eV_mu = 1.23984186
    #vacimp = 376.7303135 #Vacuum impedance in Ohms
    #eev = 510998.902 #Energy units (mc^2) in eV

    photEn_xlamds = light_eV_mu*1.e-06/xlamds

    w0 = sqrt(zrayl*xlamds/pi)
    sigrB = sqrt(rxbeam*rxbeam + rybeam*rybeam)
    #sigrF = w0
    dGrid = 0.5*rmax0*(sigrB + w0) #Mesh from -grid to +grid; This is the genesis input parameter dgrid/2

    #This defines a constant to be used for conversion (multiplication) of GENESIS Electric field
    #to SRW TD Electric Field in units of sqrt(W/mm^2)
    #xkper0 = 2*pi/xlamd #undulator "wavenumber"
    #xks = 2*pi/xlamds
    #convE_Gen2SRW = (1e-03)*xkper0*xkper0*eev/(xks*sqrt(vacimp))

    xStep = 2.*dGrid/(ncar - 1)
    convE_Gen2SRW = sqrt(1.e-06/(xStep*xStep)) #? See Genesis 1.3 manual, page 41
    #convE_Gen2SRW = 1. #?
    #print('convE_Gen2SRW=', convE_Gen2SRW)

    arEx = uti_io_read_Genesis_rad(_fpGenRad, ncar, _mult=convE_Gen2SRW, _slice_skip_per=_slice_skip_per)
    lenArE = len(arEx)
    nt = int(round(lenArE/(ncar*ncar*2)))
    arEy = None
    if(_pol == 'v'): #to treat circular polarization!?
        arEy = arEx
        arEx = None
    if(arEx == None): arEx = srwl_uti_array_alloc('f', lenArE)
    if(arEy == None): arEy = srwl_uti_array_alloc('f', lenArE)

    tStepOrig = xlamds*zsep/lightSpeed;
    tStep = tStepOrig*(_slice_skip_per + 1)
    tStartOrig = xlamds*zsep*ntail/lightSpeed;
    tStart = tStartOrig + _slice_start*tStepOrig
    tEnd = tStart + tStep*(nt - 1)

    presFT = 1 #presentation/domain: 0- frequency (photon energy), 1- time
    
    if(itdp == 0):
        tStart = photEn_xlamds #to treat harmonics!!!
        tStep = 0
        presFT = 0

    wfr = SRWLWfr(arEx, arEy, _typeE='f',
                  _eStart=tStart, _eFin=tEnd, _ne=nt,
                  _xStart=-dGrid, _xFin=dGrid, _nx=ncar,
                  _yStart=-dGrid, _yFin=dGrid, _ny=ncar, _zStart=0)
    wfr.avgPhotEn = photEn_xlamds
    wfr.presFT = presFT
    wfr.unitElFld = 2 #electric field units are sqrt(J/eV/mm^2) or sqrt(W/mm^2), depending on representation (freq. or time)

    return wfr


'''    
    n = 0
    #id = 0
    iss = 0
    
    numSlicesToReadAtOnce = 10 #to tune

    #for piece in read_in_chunks(f, size  = slice_size):
    for piece in read_in_chunks(f, size = numSlicesToReadAtOnce*slice_size):

        print('slices read:', iss)

        for iSliceRead in range(numSlicesToReadAtOnce):

            if (idx == None) or (n in idx):
                #print 'reading', n
                #if ( len(piece) / 16 != ncar**2):
                    #print 'warning, wrong slice size'
                    #print('warning, wrong slice size')
        
                #for i in xrange(len(piece) / 16 ):
                for i in range(ncarE2): #OC: for compatibility with Py3
                
                    i2 = i % _ncar
                    i1 = int(i / _ncar)
                    #print struct.unpack('d',piece[16*i:16*i+8])
                    #slices[id,i1,i2] = struct.unpack('d',piece[16*i:16*i+8])[0] + 1j*struct.unpack('d',piece[16*i+8:16*(i+1)])[0]
                    ofst = iss + i1*perX + i2*perY
                    slices[ofst] = struct.unpack('d', piece[16*i:16*i+8])[0]
                    slices[ofst + 1] = struct.unpack('d', piece[16*i+8:16*(i+1)])[0]

                    #if(slices[ofst] != 0.): print(slices[ofst], slices[ofst + 1])

                #print('slice #', iss, ' read')

                #id += 1
                iss += 1
            n += 1
    
    #print('read slices: ', slices.shape)
    return slices


def readRadiationFile_mpi(comm=None, fileName='simulation.gout.dfl', npoints=51):
    """
    not advisable to be used with very small n_proc due to memory overhead ~ file_size / n_proc  
    """
    from mpi4py import MPI
    
    def read_in_chunks(file, size=1024):
        while True:
            data = file.read(size)
            if not data:
                break
            yield data

    
    rank = comm.Get_rank()
    nproc = comm.Get_size()
        
    f = open(fileName,'rb')
    f.seek(0,2)
    total_data_len = f.tell() / 8 / 2
    f.seek(0,0)
    slice_size = int(2.0*npoints*npoints * 8.0)
    n_slices = int(total_data_len / (npoints**2))

    ncar = int(np.sqrt((slice_size/16)))
    
    local_data_len = int(n_slices / nproc)
    
    n_extra = n_slices - local_data_len * nproc
    
    tmp_buf = np.zeros([local_data_len,ncar,ncar], dtype=complex)

    
    if rank == 0:
        slice_start  = rank * local_data_len
        slices = np.zeros([n_slices,ncar,ncar], dtype=complex)
        slices_to_read = local_data_len + n_extra
    else:
        slice_start = rank * local_data_len + n_extra
        slices = []
        slices_to_read = local_data_len
        
    n = 0
    
    f.seek(slice_start*slice_size, 0)

    print 'rank', rank, ' reading', slice_start, slices_to_read, n_extra

    for piece in read_in_chunks(f, size  = slice_size):
                
        if n >= slices_to_read :
            break
        
        if ( len(piece) / 16 != ncar**2):
            print 'warning, wrong slice size'
    
        for i in xrange(len(piece) / 16 ):
            i2 = i % ncar
            i1 = int(i / ncar)
            if rank == 0:
                #print n, n_extra
                v = struct.unpack('d',piece[16*i:16*i+8])[0] + 1j*struct.unpack('d',piece[16*i+8:16*(i+1)])[0]
                slices[n,i1,i2] = v
                if n - n_extra >= 0:
                    tmp_buf[n-n_extra,i1,i2] = v
            else:
                tmp_buf[n,i1,i2] = struct.unpack('d',piece[16*i:16*i+8])[0] + 1j*struct.unpack('d',piece[16*i+8:16*(i+1)])[0] 
        n += 1
    #print rank, 'tmp_buf=', tmp_buf
    comm.Gather([tmp_buf,  MPI.COMPLEX], [slices[n_extra:], MPI.COMPLEX])
    
    return slices


#################################################################################################
####################################   End routines from Ocelot #################################
#################################################################################################

#################################################################################################
####################             Routines NOT used in prop_par.py           #####################
####################          but useful for near-future extensions         #####################
#################################################################################################

def fluence(field,Xdim,Ydim,nx,ny,Ntotph):

    Dx = Xdim/nx
    Dy = Ydim/ny
    print('Dx = ',Dx,' m')
    print('Dy = ',Dy,' m')
    integral = 0    
    maxintens = field[0]*field[0] + field[1]*field[1]    
    for j in np.arange(0,len(field),2):        
        intens = field[j]*field[j] + field[j+1]*field[j+1]
        integral = integral + intens * Dx * Dy
        if maxintens < intens: maxintens = intens        
    result = Ntotph*maxintens/integral
    print('max intensity = ',maxintens,' A.U.')
    print('integrated intensity = ',integral,' A.U.')
    print('Xdim = ',Xdim,' m')
    print('Ydim = ',Ydim,' m')
                        
    return result


def fluenceI(Inte,Xdim,Ydim,nx,ny,Ntotph):

    Dx = Xdim/nx
    Dy = Ydim/ny
    print('Dx = ',Dx,' m')
    print('Dy = ',Dy,' m')
    integral = 0    
    maxintens = Inte[0]    
    for j in np.arange(0,len(Inte),1):        
        intens = Inte[j]
        integral = integral + intens * Dx * Dy
        if maxintens < intens: maxintens = intens        
    result = Ntotph*maxintens/integral
    print('max intensity = ',maxintens,' A.U.')
    print('integrated intensity = ',integral,' A.U.')
    print('Xdim = ',Xdim,' m')
    print('Ydim = ',Ydim,' m')
                        
    return result



def writeres(namef, dataX, dataY):

    f1 = open(namef, 'w')
    for i in range(len(dataX)):
        f1.write('%s ' %(dataX[i]) + '%s' %(dataY[i]) +'\n')        
    f1.close()
    




def findsourceave(field, wfr0, zini, zfin, zstep, slicein, slicefin):
    
    reg_arzi = []
    wfrini = copy.deepcopy(wfr0)
    NX = wfrini.mesh.nx
    NY = wfrini.mesh.ny
    
    #print('nx0',nx0)
    #print('ny0',ny0)
    R_Xs0=[]
    R_Xf0=[]
    R_nx0=[]
    R_Ys0=[]
    R_Yf0=[]
    R_ny0=[]
    
    for qsl in range(slicein,slicefin):

        print('Finding the source. Processing slice nr ',qsl-slicein+1,' out of ',slicefin-slicein) 
        sl = field[qsl]
        count = 0
        for j in range(NY):
            for k in range(NX):
                wfrini.arEx[count] = np.float(np.real(sl[j][k]))
                count = count + 1
                wfrini.arEx[count] = np.float(np.imag(sl[j][k]))
                count = count + 1
        OUT = findsource(wfrini, zini, zfin, zstep)
        registerARZI = OUT[5]
        registerMESH = OUT[8]

        
        
        inte_registerARZI = []
        
        
        if qsl == slicein:
            for num in range(len(registerARZI)):
                R_Xs0.append(registerMESH[num].xStart)
                R_Xf0.append(registerMESH[num].xFin)
                R_nx0.append(registerMESH[num].nx)
                R_Ys0.append(registerMESH[num].yStart)
                R_Yf0.append(registerMESH[num].yFin)
                R_ny0.append(registerMESH[num].ny)

        for num in range(len(registerARZI)):

            Xs0 = R_Xs0[num]
            Xf0 = R_Xf0[num]
            nx0 = R_nx0[num]
            Ys0 = R_Ys0[num]
            Yf0 = R_Yf0[num]
            ny0 = R_ny0[num]
            
                  
            XMESH = np.linspace(Xs0, Xf0, nx0)
            YMESH = np.linspace(Ys0, Yf0, ny0)
            
            #print(Xs0, ' ', Xf0, ' ',nx0, ' ',ny0)                
            Xs=registerMESH[num].xStart
            Xf=registerMESH[num].xFin
            nx=registerMESH[num].nx
            Ys=registerMESH[num].yStart
            Yf=registerMESH[num].yFin
            ny=registerMESH[num].ny
            Xact = np.linspace(Xs, Xf, nx)
            Yact = np.linspace(Ys, Yf, ny)
            dataI = np.array( registerARZI[num] )
            zI    = dataI.reshape( ny, nx )
            Interdata = interpolate.interp2d(Xact, Yact, zI)
            zI_inte    = Interdata(XMESH, YMESH)
            INT_arI   = zI_inte.reshape(ny0,nx0)
            inte_registerARZI.append(np.array(INT_arI))
            #print('slice ',qsl,' len = ',len(inte_registerARZI),' with ',len(np.array(INT_arI)))


            
            #plotinte(str(qsl)+'_'+str(num)+'old_pos' ,Xs,  Xf,  nx,  Ys,  Yf,  ny,  zI)
            #plotinte(str(qsl)+'_'+str(num)+'new_pos' ,Xs0, Xf0, nx0, Ys0, Yf0, ny0, INT_arI)
            
        
        if qsl == slicein:
            reg_arzi=np.array(inte_registerARZI)
        else:
            reg_arzi = reg_arzi + np.array(inte_registerARZI)
            

    InteWaist = 0  
    registerI = []
    registerZ = OUT[1]    
    register_fwhm_X = []
    register_fwhm_Y = []


    #print(len(reg_arzi))
    #print(len(reg_arzi[0]))
    #print(len(reg_arzi[0][0]))

    #print(len(reg_arzi),'... e` 19?')
    
    InteWaist = 0
    for k in range(len(reg_arzi)):        
        intens = np.max(reg_arzi[k])
        registerI.append(intens)        
        if intens > InteWaist:
            InteWaist = intens                       
            WaistPos = OUT[1][k]
            xwaist = 1e6 * np.linspace(R_Xs0[k], R_Xf0[k], R_nx0[k] )
            ywaist = 1e6 * np.linspace(R_Ys0[k], R_Yf0[k], R_ny0[k] )
            
        x = 1e6 * np.linspace(R_Xs0[k], R_Xf0[k], R_nx0[k] )
        y = 1e6 * np.linspace(R_Ys0[k], R_Yf0[k], R_ny0[k] )
        arIpy=np.sum(reg_arzi[k],1)
        arIpyf, fwhmy=fwhm_gauss_fit(y,arIpy)        
        arIpx=np.sum(reg_arzi[k],0)
        arIpxf, fwhmx=fwhm_gauss_fit(x,arIpx)        
        register_fwhm_X.append(fwhmx)
        register_fwhm_Y.append(fwhmy) 
          
    return [registerZ,registerI,WaistPos,InteWaist,register_fwhm_X,register_fwhm_Y,reg_arzi, xwaist, ywaist]
        





def findsource(wfrini1, zini, zfin, zstep):

    Zpos = zini-zstep
    flagf    = 0
    j = 1
    registerI = []
    registerZ = []
    registerARZI = []
    register_fwhm_X = []
    register_fwhm_Y = []
    register_mesh   = []
    print('Initial wavefront position = ',wfrini1.mesh.zStart)

    while Zpos>zfin:
        
        Drift_1 = SRWLOptD(-j*zstep+zini) #Drift    
        #Wavefront Propagation Parameters with OM:
        #                    [ 0] [1] [2]  [3] [4] [5]  [6]  [7]  [8]  [9] [10] [11]         
        ppDrift_1 =          [ 1,  1, 1.0,  0,  0, 1.0, 1.0, 1.0, 1.0,  0,  0,   0]   
        ppFinal =            [ 0,  0, 1.0,  0,  0, 1.0, 1.0, 1.0, 1.0,  0,  0,   0]

        #ppDrift_1 =          [ 0,  0, 1.0,  0,  0, 1.0, 2.0, 1.0, 2.0,  0,  0,   0]   
        #ppFinal =            [ 0,  0, 1.0,  0,  0, 0.5, 0.5, 0.5, 0.5,  0,  0,   0]
        
        optBL = SRWLOptC([Drift_1],  [ppDrift_1, ppFinal])
        wfr1c = copy.deepcopy(wfrini1)                
        srwl.PropagElecField(wfr1c, optBL)
        arI = array('f', [0]*wfr1c.mesh.nx*wfr1c.mesh.ny) #"flat" array to take 2D intensity data
        srwl.CalcIntFromElecField(arI, wfr1c, 6, 0, 3, wfr1c.mesh.eStart, 0, 0) #extracts intensity        
        arP = array('d', [0]*wfr1c.mesh.nx*wfr1c.mesh.ny) #"flat" array to take 2D phase data (note it should be 'd')
        srwl.CalcIntFromElecField(arP, wfr1c, 0, 4, 3, wfr1c.mesh.eStart, 0, 0) #extracts radiation phase
        register_mesh.append(wfr1c.mesh)
        dataI = np.array( arI )
        zI=dataI.reshape( wfr1c.mesh.ny, wfr1c.mesh.nx )
        registerARZI.append(zI)
        x = 1e6 * np.linspace(wfr1c.mesh.xStart, wfr1c.mesh.xFin, wfr1c.mesh.nx )
        y = 1e6 * np.linspace(wfr1c.mesh.yStart, wfr1c.mesh.yFin, wfr1c.mesh.ny )
        arIpy=np.sum(zI,1)
        arIpyf, fwhmy=fwhm_gauss_fit(y,arIpy)        
        arIpx=np.sum(zI,0)
        arIpxf, fwhmx=fwhm_gauss_fit(x,arIpx)        
        register_fwhm_X.append(fwhmx)
        register_fwhm_Y.append(fwhmy) 
        field = wfr1c.arEx
        maxintens = field[0]*field[0] + field[1]*field[1]    
        for k in np.arange(0,len(field),2):        
            intens = field[k]*field[k] + field[k+1]*field[k+1]
            if maxintens < intens: maxintens = intens
        registerI.append(maxintens)
        registerZ.append(Zpos)           
        Zpos = Zpos - zstep
        j = j+1
        #plotinte('singlePOS'+str(j) ,wfr1c.mesh.xStart, wfr1c.mesh.xFin, wfr1c.mesh.nx,  wfr1c.mesh.yStart, wfr1c.mesh.yFin, wfr1c.mesh.ny,   zI)

    InteWaist = np.max(registerI)    
    maxind = np.argmax(registerI)
    WaistPos = registerZ[maxind]

    [wfr1d, arI, arP] = Source_Prop(Len_drft = WaistPos-zini, Wfrin = wfrini1) 

    
    return [wfr1d,registerZ,registerI,WaistPos,InteWaist,registerARZI,register_fwhm_X,register_fwhm_Y,register_mesh]




def Source_slice(wfrG=[], zin = 0.0, zfin = -40., zstep=1.0):

    # ****************************** Finds the source position for slice slicen ************************************

    OUT    = findsource(wfrG, zin, zfin, zstep)
    registerZ       = OUT[1]
    registerI       = OUT[2]
    register_fwhm_X = OUT[6]
    register_fwhm_Y = OUT[7]
    writeres(res_dir+'IvsZ_sl'+str(slicen+1)+'.dat',registerZ,registerI)
    writeres(res_dir+'fwhmXvsZ_sl'+str(slicen+1)+'.dat',registerZ,register_fwhm_X)
    writeres(res_dir+'fwhmYvsZ_sl'+str(slicen+1)+'.dat',registerZ,register_fwhm_Y)

    WfrGTempl = OUT[0]

    arIP = array('f', [0]*WfrGTempl.mesh.nx*WfrGTempl.mesh.ny) #"flat" array to take 2D intensity data
    srwl.CalcIntFromElecField(arIP, WfrGTempl, 6, 0, 3, WfrGTempl.mesh.eStart, 0, 0) #extracts intensity    
    arPP = array('d', [0]*WfrGTempl.mesh.nx*WfrGTempl.mesh.ny) #"flat" array to take 2D phase data (note it should be 'd')
    srwl.CalcIntFromElecField(arPP, WfrGTempl, 0, 4, 3, WfrGTempl.mesh.eStart, 0, 0) 
    plotfield("Source_sl"+str(sample+1), WfrGTempl.mesh.xStart, WfrGTempl.mesh.xFin, WfrGTempl.mesh.nx, WfrGTempl.mesh.yStart, WfrGTempl.mesh.yFin, WfrGTempl.mesh.ny, arIP, arPP)

  


def xyprofave(pulse, tgrid, slicein, slicefin):
   
    print('Finding the average profile')
    for j in range(slicein,slicefin):
        slices = pulse[j]
        OUT = xyprofile(slices,tgrid)
        if j == slicein :
            Xcutave = OUT[0][1]
            Ycutave = OUT[1][1]
        else :
            Xcutave = Xcutave + OUT[0][1]
            Ycutave = Ycutave + OUT[1][1]
    Xcoord = OUT[0][0]
    Ycoord = OUT[1][0]
    Xcutave = Xcutave/(slicefin-slicein)
    Ycutave = Ycutave/(slicefin-slicein)

    return [[Xcoord, Xcutave],[Ycoord, Ycutave]]





def xydivave(pulse, tgrid, slicein, slicefin, lambd):
   
    print('Finding the average divergence')
    for j in range(slicein,slicefin):
        slices = pulse[j]
        OUT = xydiv(slices,tgrid,lambd)
        if j == slicein :
            Xcutave = OUT[0][1]
            Ycutave = OUT[1][1]
        else :
            Xcutave = Xcutave + OUT[0][1]
            Ycutave = Ycutave + OUT[1][1]
    Xcoord = OUT[0][0]
    Ycoord = OUT[1][0]
    Xcutave = Xcutave/(slicefin-slicein)
    Ycutave = Ycutave/(slicefin-slicein)

    return [[Xcoord, Xcutave],[Ycoord, Ycutave]]





def xyprofile(slices, tgrid):

    Xpos = int(np.argmax(np.abs(slices))/len(slices))
    Ypos = len(slices) -1 - int(np.argmax(np.abs(slices))/len(slices))
    ##print('max = ',np.max(np.abs(slices)),' maxpos = ',np.argmax(slices))
    ##print('XPOS = ',Xpos,' YPOS = ',Ypos,' MIDDLE = ',int(len(slices)/2))
    ##print('slmax = ',slices[Xpos,Ypos],' ',slices[Ypos,Xpos])
    #Xcut = np.abs(slices[int(len(slices)/2),:])**2
    #Ycut = np.abs(slices[:,int(len(slices)/2)])**2
    Xcut = np.abs(slices[Xpos,:])**2
    Ycut = np.abs(slices[:,Ypos])**2
    Xcoord = np.linspace(-tgrid,tgrid,len(Xcut))
    Ycoord = np.linspace(-tgrid,tgrid,len(Xcut))    
    #matplotlib.pyplot.figure()
    #matplotlib.pyplot.plot(Xcoord,Xcut)
    #matplotlib.pyplot.show()
    #matplotlib.pyplot.figure()
    #matplotlib.pyplot.plot(Ycoord,Ycut)
    #matplotlib.pyplot.show()

    return [[Xcoord, Xcut],[Ycoord, Ycut]]





def xydiv(slices, tgrid, lambd):    

    fslices = np.fft.fftshift(np.fft.fft2(slices))

    KXpos = int(np.argmax(np.abs(fslices))/len(fslices))
    KYpos = len(fslices) - 1 - int(np.argmax(np.abs(fslices))/len(fslices))
    ##print('KXPOS = ',KXpos,' KYPOS = ',KYpos,' MIDDLE = ',int(len(fslices)/2))
                  
    #KXcut = np.abs(fslices[int(len(slices)/2),:])**2
    #KYcut = np.abs(fslices[:,int(len(slices)/2)])**2
    KXcut = np.abs(fslices[KXpos,:])**2
    KYcut = np.abs(fslices[:,KYpos])**2  
    KXmax = 2*pi/(2*tgrid/len(KXcut))/2.0 #Total range is 2pi/(smallest size); smallest size is 2tgrid/len(KXcut); Max is total range/2
    KYmax = 2*pi/(2*tgrid/len(KYcut))/2.0
    THXmax = KXmax * lambd/(2*np.pi)
    THYmax = KYmax * lambd/(2*np.pi)        
    THXcoord = np.linspace(-THXmax,THXmax,len(KXcut))
    THYcoord = np.linspace(-THYmax,THYmax,len(KYcut))    
    #matplotlib.pyplot.figure()
    #matplotlib.pyplot.plot(THXcoord,KXcut)
    #matplotlib.pyplot.show()
    #matplotlib.pyplot.figure()
    #matplotlib.pyplot.plot(THYcoord,KYcut)
    #matplotlib.pyplot.show()
    
    return [[THXcoord, KXcut],[THYcoord, KYcut]]



def ProDiv_slice(sl = [], dgrid = 1.0, xlamds = 1.0):

    # ****************************** Calculates profile and divergence of slice slicen ************************************

    xy   = xyprofile(sl,dgrid)
    writeres(res_dir+'X_prof_sl'+str(slicen+1)+'.dat',xy[0][0],xy[0][1])
    writeres(res_dir+'Y_prof_sl'+str(slicen+1)+'.dat',xy[1][0],xy[1][1])
    matplotlib.pyplot.figure()
    matplotlib.pyplot.plot(xy[0][0],xy[0][1])
    savefig(res_dir+'X_prof_sl'+str(slicen+1)+'.png')
    matplotlib.pyplot.figure()
    matplotlib.pyplot.plot(xy[1][0],xy[1][1])
    savefig(res_dir+'Y_prof_sl'+str(slicen+1)+'.png')

    thxy = xydiv(sl,dgrid,xlamds)
    writeres(res_dir+'X_div_sl'+str(slicen+1)+'.dat',thxy[0][0],thxy[0][1])
    writeres(res_dir+'Y_div_sl'+str(slicen+1)+'.dat',thxy[1][0],thxy[1][1])
    matplotlib.pyplot.figure()
    matplotlib.pyplot.plot(thxy[0][0],thxy[0][1])
    savefig(res_dir+'X_div_sl'+str(slicen+1)+'.png')
    matplotlib.pyplot.figure()
    matplotlib.pyplot.plot(thxy[1][0],thxy[1][1])
    savefig(res_dir+'Y_div_sl'+str(slicen+1)+'.png')

def GenWfr_slice(wfrG = []):

    # **********************Calculating Genesis Wavefront and extracting Intensity for slice slicen:

    XsG=wfrG.mesh.xStart
    XfG=wfrG.mesh.xFin
    nxG=wfrG.mesh.nx
    YsG=wfrG.mesh.yStart
    YfG=wfrG.mesh.yFin
    nyG=wfrG.mesh.ny

    arIG = array('f', [0]*nxG*nyG) #"flat" array to take 2D intensity data
    srwl.CalcIntFromElecField(arIG, wfrG, 6, 0, 3, wfrG.mesh.eStart, 0, 0) #extracts intensity
    arPG = array('d', [0]*nxG*nyG) #"flat" array to take 2D phase data (note it should be 'd')
    srwl.CalcIntFromElecField(arPG, wfrG, 0, 4, 3, wfrG.mesh.eStart, 0, 0) #extracts radiation phase
    plotfield("Gen_out_sl"+str(slicen+1), XsG, XfG, nxG, YsG, YfG, nyG, arIG, arPG)



def AveSourcePos(field = [], wfrG=[], zin = -20.0, zfin = -40.0, step = 1.0, slicein=-20.0, slicefin=-40.0):


    # ****************************** Finds the average source position ************************************

    OUTAVE = findsourceave(field, wfrG, -20.0, -40.0, 1.0, slicein, slicefin)
    registerZ_ave       = OUTAVE[0]
    registerI_ave       = OUTAVE[1]
    register_fwhm_X_ave = OUTAVE[4]
    register_fwhm_Y_ave = OUTAVE[5]
    reg_arzi_ave        = OUTAVE[6]
    writeres(res_dir+'IvsZ_ave.dat',registerZ_ave,registerI_ave)
    writeres(res_dir+'fwhmXvsZ_ave.dat',registerZ_ave,register_fwhm_X_ave)
    writeres(res_dir+'fwhmYvsZ_ave.dat',registerZ_ave,register_fwhm_Y_ave)
    maxind = np.argmax(registerI_ave)

    x = OUTAVE[7]
    y = OUTAVE[8]
    arIpy=np.sum(reg_arzi_ave[maxind],1)
    arIpx=np.sum(reg_arzi_ave[maxind],0)
    #These are summed! The previous ones are cut 
    writeres(res_dir+'IvsX_ave.dat',x,arIpx)
    writeres(res_dir+'IvsY_ave.dat',y,arIpy)

    return OUTAVE[2]   


      
def prtfluence(wfr = [], arI=[], MESH = [] ):

    Xs = MESH[0]
    Xf = MESH[1]
    nx = MESH[2]
    Ys = MESH[3]
    Yf = MESH[4]        
    ny = MESH[5]
    pixel_area=(Xf-Xs)*(Yf-Ys)/nx/ny
    integral=sum([x*pixel_area for x in arI])
    np.disp('integral_area(a.u.)= '+str(integral/1000000))

    np.disp('max_intensity= '+str(round_sig(max(arI),3)))
    np.disp('pixel_area= '+str(pixel_area))
    Ntotph = 1
    flux = fluence(wfr.arEx, Xf-Xs, Yf-Ys, nx, ny, 1)
    print('X size = ',Xf-Xs,' m')
    print('Y size = ',Yf-Ys,' m')
    print('X mesh = ',nx,' points')
    print('Y mesh = ',ny,' points')
    print('Total fluence is ',flux,'*Nph in ph/m2')
    print('also corresponding to ',flux/1.0e6,'*Nph in ph/mm2')


def AVE_prtfluence(INT_arI_ave_P = [], MESH = []):

    XsP0 = MESH[0]
    XfP0 = MESH[1]
    nxP0 = MESH[2]
    YsP0 = MESH[3]
    YfP0 = MESH[4]        
    nyP0 = MESH[5]
    pixel_area=(XfP0-XsP0)*(YfP0-YsP0)/nxP0/nyP0
    integral=sum([x*pixel_area for x in INT_arI_ave_P])    
    np.disp('integral_area(a.u.)= '+str(integral/1000000))

    np.disp('max_intensity= '+str(round_sig(max(INT_arI_ave_P),3)))
    np.disp('pixel_area= '+str(pixel_area))
    Ntotph = 1
    flux = fluenceI(INT_arI_ave_P, XfP0-XsP0, YfP0-YsP0, nxP0, nyP0, 1)
    print('X size = ',XfP0-XsP0,' m')
    print('Y size = ',YfP0-YsP0,' m')
    print('X mesh = ',nxP0,' points')
    print('Y mesh = ',nyP0,' points')
    print('Total fluence is ',flux,' ph/m2')
    print('also corresponding to ',flux/1.0e6,' ph/mm2')


'''

'''
#################################################################################################
####################           Routines strictly used in prop_par.py        #####################
#################################################################################################

def GeneRead(comm=[], filename='D:\prog\data\PROP\sase2.4p1keV.tap.1.out.dfl', Tmesh=301, PhEne = 4100.0, dgrid = 1.0):
    """
    Reads the Genesis file and gives back wavefront backbone and field as a np.array
    """
    
    
    wfrG = SRWLWfr()               #Initial Electric Field Wavefront
    wfrG.allocate(1, Tmesh, Tmesh) #Numbers of points vs Photon Energy (1), Horizontal and Vertical Positions (dummy)
    wfrG.mesh.zStart =  0.0        #Longitudinal Position [m] at which Electric Field has to be calculated, i.e. the position of the first optical element
    wfrG.mesh.eStart =  PhEne      #Initial Photon Energy [eV]
    wfrG.mesh.eFin   =  PhEne      #Final Photon Energy [eV]
    wfrG.mesh.xStart = -dgrid      #Initial Horizontal Position [m]
    wfrG.mesh.xFin   =  dgrid      #Final Horizontal Position [m]
    wfrG.mesh.yStart = -dgrid      #Initial Vertical Position [m]
    wfrG.mesh.yFin   =  dgrid      #Final Vertical Position [m]

    wfrG.partBeam.partStatMom1.x  = 0.0     #Some information about the source in the Wavefront structure
    wfrG.partBeam.partStatMom1.y  = 0.0
    wfrG.partBeam.partStatMom1.z  = 0.0
    wfrG.partBeam.partStatMom1.xp = 0.0
    wfrG.partBeam.partStatMom1.yp = 0.0

    #Reading Genesis field file
    print('Reading Genesis field file...')
    field=readRadiationFile_mpi(comm=comm, fileName=filename, npoints=Tmesh)
    print('...Done.',rank,'. ',len(field))
    
    return [wfrG, field]


def Source_Prop(Len_drft = 0.1, Wfrin = []):
    
    #**********************Calculating virtual Source Wavefront and extracting Intensity:

    Drift_1 = SRWLOptD(Len_drft) #Drift    
    # 2colors U2 sase3 :
    #                    [ 0] [1] [2]  [3] [4] [5]  [6]  [7]  [8]  [9] [10] [11] 
    ppDrift_1 =          [ 0,  0, 1.0,  0,  0, 1.0, 2.0, 1.0, 2.0,  0,  0,   0]   
    ppFinal =            [ 0,  0, 1.0,  0,  0, 0.5, 1.0, 0.5, 1.0,  0,  0,   0]

    # Usual source sase1
    #ppDrift_1 =          [ 0,  0, 1.0,  0,  0, 1.0, 2.0, 1.0, 2.0,  0,  0,   0]   
    #ppFinal =            [ 0,  0, 1.0,  0,  0, 0.25, 2.0, 0.25, 2.0,  0,  0,   0]
   
    
    #2colors_backUndu sase3
    #ppDrift_1 =          [ 0,  0, 1.0,  0,  0, 16.0, 1.0, 16.0, 1.0,  0,  0,   0]   
    #ppFinal =            [ 0,  0, 1.0,  0,  0, 1.0, 0.1, 1.0, 0.1,  0,  0,   0]
    
    
    optBL = SRWLOptC([Drift_1],  [ppDrift_1, ppFinal])
    wfr1d = copy.deepcopy(Wfrin)
    srwl.PropagElecField(wfr1d, optBL)
    arI = array('f', [0]*wfr1d.mesh.nx*wfr1d.mesh.ny) #"flat" array to take 2D intensity data
    srwl.CalcIntFromElecField(arI, wfr1d, 6, 0, 3, wfr1d.mesh.eStart, 0, 0) #extracts intensity    
    arP = array('d', [0]*wfr1d.mesh.nx*wfr1d.mesh.ny) #"flat" array to take 2D phase data (note it should be 'd')
    srwl.CalcIntFromElecField(arP, wfr1d, 0, 4, 3, wfr1d.mesh.eStart, 0, 0) #extracts radiation phase

    return [wfr1d, arI, arP]
    
    

def FullProp(zprop = 933.0, wfrS = [], line = SASE1_SPB_line, ppFinal = [ 0,  0, 1.0,  1,  0, 1.0, 1.0, 1.0, 1.0,  0,  0,   0]):
    
    print line
    [ Act_Elem_box, Act_Param_box ] = line(zprop)
    Act_Param_box.append( ppFinal )
    OptBL = SRWLOptC(Act_Elem_box, Act_Param_box)
    wfrP = copy.deepcopy(wfrS) #Genesis backpropagated    
    srwl.PropagElecField(wfrP, OptBL)
    
    return wfrP
    
    
def AVE_Save_Inte(wfrP=[], MESH = [], INT_arI_ave_P = [],  firstslice = 1):

    XsP=wfrP.mesh.xStart
    XfP=wfrP.mesh.xFin
    nxP=wfrP.mesh.nx
    YsP=wfrP.mesh.yStart
    YfP=wfrP.mesh.yFin
    nyP=wfrP.mesh.ny
    
    arIP = array('f', [0]*nxP*nyP) #"flat" array to take 2D intensity data
    srwl.CalcIntFromElecField(arIP, wfrP, 6, 0, 3, wfrP.mesh.eStart, 0, 0) #extracts intensity
    if firstslice == 1:        
        INT_arI_ave_P = np.array(arIP)        
        XsP0 = XsP
        XfP0 = XfP
        YsP0 = YsP
        YfP0 = YfP
        nxP0 = nxP
        nyP0 = nyP
    else:
        XsP0 = MESH[0]
        XfP0 = MESH[1]
        nxP0 = MESH[2]
        YsP0 = MESH[3]
        YfP0 = MESH[4]        
        nyP0 = MESH[5]
        XMESH = np.linspace(XsP0, XfP0, nxP0)
        YMESH = np.linspace(YsP0, YfP0, nyP0)
        dataI = np.array( arIP )
        zI    = dataI.reshape( nyP, nxP )
        Xact = np.linspace(XsP, XfP, nxP)
        Yact = np.linspace(YsP, YfP, nyP)
        Interdata = interpolate.interp2d(Xact, Yact, zI)
        zI_inte    = Interdata(XMESH, YMESH)
        INT_arIP   = zI_inte.reshape(nyP*nxP)
        INT_arI_ave_P = INT_arI_ave_P + np.array(INT_arIP)

    wfrP.mesh.xStart = XsP0
    wfrP.mesh.xFin   = XfP0
    wfrP.mesh.nx     = nxP0
    wfrP.mesh.yStart = YsP0
    wfrP.mesh.yFin   = YfP0
    wfrP.mesh.ny     = nyP0
    wfrP.mesh.ne     = 1
    

    return [INT_arI_ave_P, [XsP0,XfP0,nxP0,YsP0,YfP0,nyP0]]


######################################
####GRATINGS ROUTINES TO BE TESTED####
######################################


def fftpulse(field_in = []):

    fft_field = field_in
    nslice = len(field_in)
    xmesh  = len(field_in[0])
    ymesh  = len(field_in[0,0])

    for j in range(xmesh):
        for k in range(ymesh):
            fft_field[:,j,k] = np.fft.fft(field_in[:,j,k])    

    return fft_field

def ifftpulse(field_in = []):

    ifft_field = field_in
    nslice = len(field_in)
    xmesh  = len(field_in[0])
    ymesh  = len(field_in[0,0])

    for j in range(xmesh):
        for k in range(ymesh):
            ifft_field[:,j,k] = np.fft.ifft(field_in[:,j,k])    


    return ifft_field

    

def grating(field_in = [], leng = 1, xlamds = 0.0, z1 = 0.0, z2 = 0.0, r_t = 0.0, r_s = 0.0, theta_i = 0.0, theta_d =0.0, dtheta_d, d0 = 0.0, d1 = 0.0, d2 = 0.0, aberration = False):
    """
    % field_in is the input complex field
    % field_out is the output complex field
    % leng is transverse size of the mesh (dgrid*2)
    % xlamds is wavelength
    % z1 and z2 are distances to source and image in order to calculate 
    %   aberration effect (we use semi-analytical approach to propagate, 
    %   so we need to provide this information)
    % R_t is tangential radius of curvature
    % R_s is sagittal radius of curvature
    % Teta_i and Teta_d are incidence and reflection angles (obviously one can 
    %   calculate it now internaly within function. I keep dummy input variable
    %   to save time on revriting all versiong of script code.
    % dTheta_d is angle deviation of the field from principle ray direction
    %   after grating. It is used to introduce different angles of reflected
    %   field from grating at different  wavelengths if time-dependent (3D) field
    %   is propagated, resulting in pusle-front tilt.
    % D0, D1, D2 are grating coefficients (in out case 1123[l/mm] 1.6[l/mm2] 0.002[l/mm3])
    % aberration is logical value, that defines whether include aberration effects, or not
    """

    
    theta_d  = np.arccos(np.cos(theta_i)-xlamds*d0*1.0e3)
    k        = 2.0*np.pi/xlamds
    gr_asym  = theta_i/theta_d         # grating asymmetry factor GG: assumes grazing angles sin angle ~ angle (?)
    gr_sigma = 1.0/(d0*1.0e3)          # sigma parameter of grating (1/k)
    gr_alpha = d1*1.0e6/(d0*1.0e3)**2  # alpha parameter of grating (old=1,8) (n1/k/k) !!
    d2       = d2*1.0e9                # in 1/m**3
    if d2 == 0.0:
        n=0
        d2=1.0
    else
        n=1

    # calculation of the sagittal focal distance 
    f_s = r_s/( np.sin(theta_i) + np.sin(theta_d) )
    #print 'f_g_s = ', f_s

    # calculation of the tangential focal distance 
    #1. vls contribution
    f_t_vls   = gr_sigma**2 * np.sin(theta_d)**2/xlamds/gr_alpha  #!!! original!!!!
    f_g_t_vls = gr_sigma**2/xlamds/gr_alpha
    #2. curvature contribution
    f_t_c     = r_t/(1.0/theta_d+theta_i/theta_d**2)              #tangential focal distance due to curvature
    #3. final tangential focal distance
    f_t       = 1.0 / ( 1.0/f_t_vls + 1.0/f_t_c )    
    #print 'f_g_t = ',f_t,' m,  f_g_t_c = ',f_t_c,' m,  f_g_t_vls = ',f_t_vls,' m'

    m      = len(fiel_in)
    dx     = leng/m
    dy     = dx
    x1     = np.linspace( (m-1.0)/2.0*dx, ((m-1.0)/2.0 + 1-m))*dx, m )
    y1     = np.linspace( (m-1.0)/2.0*dy, ((m-1.0)/2.0 + 1-m))*dy, m )
    xx, yy = np.meshgrid(x1, y1) 

    #c_20=(theta_i**2/z1 + theta_i/r_t + theta_d**2/z2 + theta_d/r_t)/2
    if aberration == True:
        c_30 = ( n*np.pi*2.0/K/3.0*d2 ) + ( np.sin(theta_i)**2/z1 - np.sin(theta_i)/r_t ) * np.cos(theta_i)/2.0/z1 - ( np.sin(theta_d)**2/z2 - sin(theta_d)/r_t )*cos(theta_d)/2.0/z2
        c_12 = ( -theta_i/r_s/z1 + 1.0/z1**2 + theta_d/r_s/z2 - 1/z2**2)/2.0
    else
        c_30 = 0
        c_12 = 0
     

    # enegry conservation calculations
    dm   = np.round( m*(1.0-gr_asym)/2.0 )
    p_in = np.sum(np.abs(field_in(dm:m-dm))**2)
    
    fx    = interpolate.interp2d(xx, yy, field_in)
    x     = fx(xx, yy*gr_asym)
    p_out = np.sum(np.abs(x)**2)
    
    if pout ~= 0.0:
        x = x*np.sqrt(p_in/p_out)
    
    x = x * np.exp(-1j*k/f_s/2.0*(xx**2))                #focusing sag
    x = x * np.exp(-1j*k/f_t/2.0*(yy**2))                #focusing tang
    x = x * np.exp(1j*k*c_30*(yy**33)/theta_d**3)        #aberration c_30 
    x = x.* np.exp(1j*k*c_12*xx**2 * yy/theta_d)         #aberration c_12

    x = np.abs(x)*np.exp(1j*(np.angle(x)+dtheta_d*k*yy)) #phase tilt
        
    field_out=x;
        
    return field_out
'''
